/* See the end of this file for copyright, license, and warranty information. */

#pragma once

#include <ardix/types.h>
#include <toolchain.h>

#define ATOM_DEFINE(val) { ._val = val, }
#define ATOM(name, val) atom_t name = ATOM_DEFINE(val)

static __always_inline void atom_init(atom_t *atom, int val)
{
	atom->_val = val;
}

static __always_inline int atom_read(const atom_t *atom)
{
	return atom->_val;
}

/*
 * These are implemented in arch/<arch>/atom.S
 */

extern int _atom_add(volatile int *atom, int val);
extern int _atom_sub(volatile int *atom, int val);
extern int _atom_and(volatile int *atom, int val);
extern int _atom_or(volatile int *atom, int val);
extern int _atom_xor(volatile int *atom, int val);
extern int _atom_xchg(volatile int *atom, int val);
extern int _atom_cmpxchg(volatile int *atom, int cmp, int val);

/**
 * @brief Atomically add `val` to `atom`.
 * @return The old value of `atom`, before the addition
 */
static __always_inline int atom_add(atom_t *atom, int val)
{
	return _atom_add(&atom->_val, val);
}

/**
 * @brief Atomically subtract `val` from `atom`.
 * @return The old value of `atom` before the subtraction
 */
static __always_inline int atom_sub(atom_t *atom, int val)
{
	return _atom_sub(&atom->_val, val);
}

/**
 * @brief Atomically do a bitwise AND of `val` and `atom`.
 * @return The old value of `atom`, before the AND
 */
static __always_inline int atom_and(atom_t *atom, int val)
{
	return _atom_and(&atom->_val, val);
}

/**
 * @brief Atomically do a bitwise OR of `val` and `atom`.
 * @return The old value of `atom`, before the OR
 */
static __always_inline int atom_or(atom_t *atom, int val)
{
	return _atom_or(&atom->_val, val);
}

/**
 * @brief Atomically do a bitwise XOR of `val` and `atom`.
 * @return The old value of `atom`, before the XOR
 */
static __always_inline int atom_xor(atom_t *atom, int val)
{
	return _atom_xor(&atom->_val, val);
}

/**
 * @brief Atomically increment `atom` by 1.
 * @return The old value of `atom`, before the increment
 */
static __always_inline int atom_get(atom_t *atom)
{
	return _atom_add(&atom->_val, 1);
}

/**
 * @brief Atomically decrement `atom` by 1.
 * @return The old value of `atom`, before the decrement
 */
static __always_inline int atom_put(atom_t *atom)
{
	return _atom_sub(&atom->_val, 1);
}

/**
 * @brief Atomically exchange the value of `atom` with `val`.
 * @return The old value of `atom`
 */
static __always_inline int atom_xchg(atom_t *atom, int val)
{
	return _atom_xchg(&atom->_val, val);
}

/**
 * @brief Atomically compare the value of `atom` with `cmp` and,
 *	if found to be equal, exchange it with `val`.
 * @param atom atom to perform the operation on
 * @param cmp value to compare the atom with
 * @param val new value to be written to atom if it is equal to `cmp`
 * @return The old value of `atom`
 */
static __always_inline int atom_cmpxchg(atom_t *atom, int cmp, int val)
{
	return _atom_cmpxchg(&atom->_val, cmp, val);
}

/*
 * This file is part of Ardix.
 * Copyright (c) 2020, 2021, 2022 Felix Kopp <owo@fef.moe>.
 *
 * Ardix is non-violent software: you may only use, redistribute,
 * and/or modify it under the terms of the CNPLv6+ as found in
 * the LICENSE file in the source code root directory or at
 * <https://git.pixie.town/thufie/CNPL>.
 *
 * Ardix comes with ABSOLUTELY NO WARRANTY, to the extent
 * permitted by applicable law.  See the CNPLv6+ for details.
 */
